from __future__ import annotations

import typing as t
import logging
import threading

from rich.console import Console

from . import now

rich_console = Console(soft_wrap=True)

_min_level = logging.ERROR
_min_level_lock = threading.Lock()

CRITICAL = logging.CRITICAL
FATAL = CRITICAL
ERROR = logging.ERROR
WARNING = logging.WARNING
WARN = WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
TRACE = 5


def set_level(level: int) -> None:
    global _min_level
    with _min_level_lock:
        if level not in _levels:
            raise ValueError(f"invalid log level: {level}")
        _min_level = level


class _LevelInfo(t.NamedTuple):
    number: int
    name: str
    color: str
    emoji: str
    style: str


_levels: t.Dict[int, _LevelInfo] = {
    TRACE: _LevelInfo(TRACE, "TRACE", "white", ":bird:", "default"),
    DEBUG: _LevelInfo(DEBUG, "DEBUG", "blue", ":speaker:", "default"),
    INFO: _LevelInfo(INFO, "INFO", "green", ":bulb:", "default"),
    WARN: _LevelInfo(WARN, "WARN", "yellow", ":question:", "bold magenta"),
    ERROR: _LevelInfo(ERROR, "ERROR", "red", ":x:", "bold red"),
    FATAL: _LevelInfo(
        FATAL,
        "FATAL",
        "bold red",
        ":x:",
        "red on white",
    ),
}


def get_level_name(level: int) -> str:
    if level not in _levels:
        return f"LEVEL {level}"
    else:
        return _levels[level].name


def print(*args: t.Any, **kwargs: t.Any) -> None:
    extend_args = list(args or [])
    global _min_level
    without_timestamp = kwargs.pop("without_timestamp", False)
    if _min_level <= logging.DEBUG and not without_timestamp:
        extend_args.insert(0, f"[{_get_datetime_str()}]")

    rich_console.print(*extend_args, **kwargs)


def print_exception(*args: t.Any, **kwargs: t.Any) -> None:
    rich_console.print_exception(*args, **kwargs)


def trace(*args: t.Any) -> None:
    _log(TRACE, *args)


def debug(*args: t.Any) -> None:
    _log(logging.DEBUG, *args)


def info(*args: t.Any) -> None:
    _log(logging.INFO, *args)


def warn(*args: t.Any) -> None:
    _log(logging.WARN, *args)


warning = warn


def error(*args: t.Any) -> None:
    _log(logging.ERROR, *args)


def exception(*args: t.Any) -> None:
    _log(logging.ERROR, *args)
    print_exception()


def fatal(*args: t.Any) -> None:
    _log(logging.FATAL, *args)


def log(*args: t.Any) -> None:
    _log(logging.INFO, *args)


def _log(level: int, *args: t.Any) -> None:
    level_info = _levels.get(level) or _levels[logging.INFO]

    global _min_level
    if level_info.number < _min_level:
        return

    datetime_str = _get_datetime_str()
    rich_console.print(
        f"[{datetime_str}] {level_info.emoji} [{level_info.color}]|{level_info.name}|[/]",
        *args,
        style=level_info.style,
    )


def rule(*args: t.Any, **kwargs: t.Any) -> None:
    rich_console.rule(*args, **kwargs)


def _get_datetime_str() -> str:
    global _min_level
    if _min_level <= logging.DEBUG:
        return now("%Y-%m-%d %H:%M:%S.%f")
    else:
        return now("%H:%M:%S")



import os

from rich import traceback

from starwhale.utils import console
from starwhale.consts import (
    ENV_LOG_LEVEL,
    ENV_LOG_VERBOSE_COUNT,
    ENV_DISABLE_PROGRESS_BAR,
)


def init_logger(verbose: int) -> None:
    if verbose == 0:
        lvl = console.ERROR
    elif verbose == 1:
        lvl = console.WARNING
    elif verbose == 2:
        lvl = console.INFO
    elif verbose == 3:
        lvl = console.DEBUG
    else:
        lvl = console.TRACE

    console.set_level(lvl)
    lvl_name = console.get_level_name(lvl)
    os.environ[ENV_LOG_LEVEL] = lvl_name
    os.environ[ENV_LOG_VERBOSE_COUNT] = str(verbose)

    if verbose > 0:
        os.environ[ENV_DISABLE_PROGRESS_BAR] = "1"
        console.print(f":space_invader: verbosity: {verbose}, log level: {lvl_name}")

    # TODO: custom debug for tb install
    traceback.install(
        console=console.rich_console, show_locals=True, max_frames=1, width=200
    )



from __future__ import annotations

import sys
import typing as t
import inspect
import numbers
import threading
from pathlib import Path
from collections import defaultdict

import yaml

from starwhale.utils import console
from starwhale.consts import DecoratorInjectAttr
from starwhale.utils.fs import ensure_file
from starwhale.base.mixin import ASDictMixin
from starwhale.utils.load import load_module
from starwhale.utils.error import NoSupportError
from starwhale.api._impl.evaluation import PipelineHandler


class Handler(ASDictMixin):
    _registered_functions: t.Dict[str, t.Callable] = {}
    _registered_handlers: t.Dict[str, Handler] = {}
    _registering_lock = threading.Lock()

    def __init__(
        self,
        name: str,
        show_name: str,
        func_name: str,
        module_name: str,
        cls_name: str = "",
        resources: t.Optional[t.Dict] = None,
        needs: t.Optional[t.List[str]] = None,
        concurrency: int = 1,
        replicas: int = 1,
        extra_args: t.Optional[t.List] = None,
        extra_kwargs: t.Optional[t.Dict] = None,
        **kw: t.Any,
    ) -> None:
        self.name = name
        self.show_name = show_name
        self.func_name = func_name
        self.module_name = module_name
        self.cls_name = cls_name
        self.resources = self._transform_resource(resources)
        self.needs = needs or []
        self.concurrency = concurrency
        self.replicas = replicas
        self.extra_args = extra_args or []
        self.extra_kwargs = extra_kwargs or {}

    def __str__(self) -> str:
        return f"Handler[{self.name}]: name-{self.show_name}"

    __repr__ = __str__

    def _transform_resource(
        self, resources: t.Optional[t.Dict[str, t.Any]]
    ) -> t.List[t.Dict]:
        resources = resources or {}
        attribute_names = ["request", "limit"]
        resource_names: t.Dict[str, t.List] = {
            "cpu": [int, float],
            "nvidia.com/gpu": [int],
            "memory": [int, float],
        }

        results = []
        for _name, _resource in resources.items():
            if _name not in resource_names:
                raise RuntimeError(
                    f"resources name is illegal, name must in {resource_names.keys()}"
                )

            if isinstance(_resource, numbers.Number):
                resources[_name] = {"request": _resource, "limit": _resource}
            elif isinstance(_resource, dict):
                if not all(n in attribute_names for n in _resource):
                    raise RuntimeError(
                        f"resources value is illegal, attribute's name must in {attribute_names}"
                    )
            else:
                raise RuntimeError(
                    "resources value is illegal, attribute's type must be number or dict"
                )

            for _k, _v in resources[_name].items():
                if type(_v) not in resource_names[_name]:
                    raise RuntimeError(
                        f"resource:{_name} only support type:{resource_names[_name]}, but now is {type(_v)}"
                    )
                if _v <= 0:
                    raise RuntimeError(
                        f"{_k} only supports non-negative number, but now is {_v}"
                    )
            results.append(
                {
                    "type": _name,
                    "request": resources[_name]["request"],
                    "limit": resources[_name]["limit"],
                }
            )
        return results

    @classmethod
    def clear_registered_handlers(cls) -> None:
        with cls._registering_lock:
            cls._registered_handlers.clear()
            cls._registered_functions.clear()

    @classmethod
    def register(
        cls,
        resources: t.Optional[t.Dict[str, t.Any]] = None,
        concurrency: int = 1,
        replicas: int = 1,
        needs: t.Optional[t.List[t.Callable]] = None,
        extra_args: t.Optional[t.List] = None,
        extra_kwargs: t.Optional[t.Dict] = None,
        name: str = "",
    ) -> t.Callable:
        """Register a function as a handler. Enable the function execute by needs handler, run with gpu/cpu/mem resources in server side,
        and control concurrency and replicas of handler run.

        Args:
            resources: [Dict, optional] Resources for the handler run, such as memory, gpu etc. Current only supports
              the cloud instance.
            concurrency: [int, optional] The concurrency of the handler run. Default is 1.
            replicas: [int, optional] The number of the handler run. Default is 1.
            needs: [List[Callable], optional] The list of the functions that need to be executed before the handler function.
              The depends callable objects must be decorated by `@handler`, `@evaluation.predict`, `@evaluation.evaluate` and `@experiment.fine_tune` .
            name: [str, optional] The user-friendly name of the handler. Default is the function name.

        Example:
        ```python
        from starwhale import handler

        @handler(resources={"cpu": 1, "nvidia.com/gpu": 1}, concurrency=2, replicas=3)
        def my_handler():
            ...

        @handler(needs=[my_handler])
        def my_another_handler():
            ...
        ```

        Returns:
            [Callable] The decorator function.
        """

        def decorator(func: t.Callable) -> t.Callable:
            if not inspect.isfunction(func):
                raise NoSupportError(
                    f"handler decorator only supports on function: {func}"
                )

            qualname = func.__qualname__
            cls_name, _, func_name = qualname.rpartition(".")
            if "." in cls_name:
                raise NoSupportError(
                    f"handler decorator no supports inner class method:{qualname}"
                )

            key_name = f"{func.__module__}:{qualname}"
            key_name_needs = []
            for n in needs or []:
                # TODO: support class as needs
                if not inspect.isfunction(n):
                    raise NoSupportError(
                        f"handler decorator no supports non-function needs:{n}"
                    )

                key_name_needs.append(f"{n.__module__}:{n.__qualname__}")

            _handler = cls(
                name=key_name,
                show_name=name or func_name,
                func_name=func_name,
                module_name=func.__module__,
                cls_name=cls_name,
                concurrency=concurrency,
                replicas=replicas,
                needs=key_name_needs,
                resources=resources,
                extra_args=extra_args,
                extra_kwargs=extra_kwargs,
            )

            with cls._registering_lock:
                cls._registered_handlers[key_name] = _handler
                cls._registered_functions[key_name] = func

            setattr(func, DecoratorInjectAttr.Step, True)
            return func

        return decorator

    @classmethod
    def get_registered_handlers_with_expanded_needs(
        cls, search_modules: t.List[str], package_dir: Path
    ) -> t.Dict[str, t.List[Handler]]:
        cls._preload_registering_handlers(search_modules, package_dir)

        with cls._registering_lock:
            expanded_names: t.Dict[str, t.Set] = {}

            for name, handler in cls._registered_handlers.items():
                if name not in expanded_names:
                    expanded_names[name] = set()

                queue = {n for n in handler.needs}

                while queue:
                    need_name = queue.pop()

                    if need_name in expanded_names[name]:
                        continue

                    if need_name == name:
                        raise RuntimeError(
                            f"cycle dependency: name-{name}, needs-{handler.needs}"
                        )

                    if need_name not in cls._registered_handlers:
                        raise RuntimeError(f"dependency not found: {need_name}")

                    expanded_names[name].add(need_name)
                    parent_need_names = (
                        expanded_names[need_name]
                        or cls._registered_handlers[need_name].needs
                    )
                    queue.update(parent_need_names)

            expanded_handlers = defaultdict(list)
            for name, need_names in expanded_names.items():
                expanded_handlers[name].extend(
                    [cls._registered_handlers[n] for n in need_names]
                )
                expanded_handlers[name].append(cls._registered_handlers[name])

            return expanded_handlers

    @classmethod
    def _preload_registering_handlers(
        cls, search_modules: t.List[str], package_dir: Path
    ) -> None:
        for module_name in search_modules:
            # handler format: a.b.c, a.b.c:d
            module_name = module_name.split(":")[0].strip()
            if not module_name or module_name == "__main__":
                continue

            # reload for the multi model.build in one python process
            if module_name in sys.modules:
                del sys.modules[module_name]
            module = load_module(module_name, package_dir)

            for v in module.__dict__.values():
                if (
                    inspect.isclass(v)
                    and issubclass(v, PipelineHandler)
                    and v != PipelineHandler
                ):
                    # compatible with old version: ppl and cmp function are renamed to predict and evaluate
                    predict_func = getattr(v, "predict", None) or getattr(v, "ppl")
                    evaluate_func = getattr(v, "evaluate", None) or getattr(v, "cmp")
                    Handler.register(replicas=1, name="predict")(predict_func)
                    Handler.register(
                        replicas=1,
                        needs=[predict_func],
                        name="evaluate",
                    )(evaluate_func)


def generate_jobs_yaml(
    search_modules: t.List[str],
    package_dir: t.Union[Path, str],
    yaml_path: t.Union[Path, str],
) -> None:
    console.print(
        f":rocket: generate jobs yaml from modules: {search_modules} , package rootdir: {package_dir}"
    )
    expanded_handlers = Handler.get_registered_handlers_with_expanded_needs(
        search_modules, Path(package_dir)
    )
    if not expanded_handlers:
        raise RuntimeError("not found any handlers")

    ensure_file(
        yaml_path,
        yaml.safe_dump(
            {
                name: [h.asdict() for h in handlers]
                for name, handlers in expanded_handlers.items()
            },
            default_flow_style=False,
        ),
        parents=True,
    )



import os
import time
import random
import string
import threading
import traceback
from typing import IO, Any, List, Union, TextIO, Callable, Optional
from datetime import datetime

import requests

from starwhale.utils import console


def random_id() -> str:
    return "".join(
        [random.choice(string.ascii_letters + string.digits) for i in range(8)]
    )


def check_status_code(resp: requests.Response) -> None:
    if resp.status_code != 200:
        msg = f"status code={resp.status_code} body={resp.json()}"
        if resp.status_code // 100 == 4:
            raise UnrecoverableError(msg)
        raise RuntimeError(msg)


class UnrecoverableError(Exception):
    pass


class InvalidOffsetError(Exception):
    pass


class ClosedError(Exception):
    pass


class TaskRunner(threading.Thread):
    def __init__(self, name: str) -> None:
        super().__init__(name=name)
        self.stopped = False
        self.daemon = True

    def stop(self) -> None:
        self.stopped = True

    def run_until_success(
        self, f: Callable[[], Any], ignore_stopped: bool = False
    ) -> Any:
        backoff = 0.1
        while not self.stopped or ignore_stopped:
            try:
                return f()
            except UnrecoverableError:
                raise
            except Exception:
                console.print("===Retry due to exception===")
                traceback.print_exc()
                console.print("======")
            time.sleep(backoff)
            backoff *= 2
            if backoff > 1:
                backoff = 1


class PeriodicRunner(TaskRunner):
    def __init__(self, name: str, f: Callable[[], None], interval: float):
        super().__init__(name)
        self.f = f
        self.interval = interval

    def run(self) -> None:
        while not self.stopped:
            try:
                start_time = datetime.now().timestamp()
                while not self.stopped:
                    now = datetime.now().timestamp()
                    elasped = now - start_time
                    if elasped > self.interval:
                        break
                    time.sleep(min(1, self.interval - elasped))
                if not self.stopped:
                    self.f()
            except Exception:
                traceback.print_exc()


class ChunkBuffer:
    def __init__(self, chunk_size: int, max_chunk_count: int):
        self.chunk_size = chunk_size
        self.max_chunk_count = max_chunk_count
        self.chunks: List[bytes] = []
        self.offset = 0
        self.length = 0
        self.lock = threading.Condition()
        self.error: Optional[str] = None

    def write(self, offset: int, data: bytes, timeout: float) -> int:
        with self.lock:
            if self.error is not None:
                raise ClosedError(self.error)
            if offset > self.length:
                raise InvalidOffsetError(f"invalid offset {offset}")
            skip_len = self.length - offset
            if skip_len >= len(data):
                return len(data)
            data = data[skip_len:]
            index = 0
            while index < len(data):
                if self.error is not None:
                    raise ClosedError(self.error)
                if len(self.chunks) == 0 or len(self.chunks[-1]) == self.chunk_size:
                    if len(self.chunks) == self.max_chunk_count:
                        if index == 0 and self.lock.wait(timeout):
                            continue
                        break
                    self.chunks.append(b"")
                    available = self.chunk_size
                else:
                    available = self.chunk_size - len(self.chunks[-1])
                if available > len(data) - index:
                    available = len(data) - index
                self.chunks[-1] += data[index : index + available]
                index += available
            self.lock.notify_all()
            self.length += index
            return index + skip_len

    def read(self, offset: int, timeout: float) -> Optional[bytes]:
        with self.lock:
            if offset < self.offset or offset > self.length:
                raise InvalidOffsetError(f"invalid offset {offset}")
            while self.error is None and offset == self.length:
                if not self.lock.wait(timeout):
                    return None
            if self.error is not None and self.error != "":
                raise ClosedError(self.error)
            if offset == self.length:
                return b""
            index = 0
            while self.offset + len(self.chunks[index]) <= offset:
                self.offset += len(self.chunks[index])
                index += 1
            if index > 0:
                self.chunks = self.chunks[index:]
                self.lock.notify_all()
            return self.chunks[0][offset - self.offset :]

    def close(self, error: str) -> None:
        with self.lock:
            if self.error is None:
                self.error = error
                self.lock.notify_all()


class ChunkBufferReader:
    def __init__(self, chunk_buffer: ChunkBuffer):
        self.chunk_buffer = chunk_buffer
        self.offset = 0

    def read(self, timeout: float) -> Optional[bytes]:
        data = self.chunk_buffer.read(self.offset, timeout)
        if data is not None:
            self.offset += len(data)
        return data

    def close(self, error: str) -> None:
        self.chunk_buffer.close(error)


class ChunkBufferWriter:
    def __init__(self, chunk_buffer: ChunkBuffer):
        self.chunk_buffer = chunk_buffer
        self.offset = 0

    def write(self, data: bytes, timeout: float) -> int:
        count = self.chunk_buffer.write(self.offset, data, timeout)
        self.offset += count
        return count

    def close(self, error: str) -> None:
        self.chunk_buffer.close(error)


class FileReader(TaskRunner):
    def __init__(
        self,
        name: str,
        file: Union[TextIO, IO[bytes]],
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.file = file
        self.chunk_buffer_writer = ChunkBufferWriter(chunk_buffer)

    def run(self) -> None:
        try:
            fd = self.file.fileno()
            while not self.stopped:
                data = os.read(fd, 16384)
                if len(data) == 0:
                    self.chunk_buffer_writer.close("")
                    return
                while not self.stopped and len(data) > 0:
                    count = self.chunk_buffer_writer.write(data, 0.5)
                    data = data[count:]
        except Exception as e:
            console.print("===Stop read from file===")
            traceback.print_exc()
            console.print("======")
            self.chunk_buffer_writer.close(str(e))
        finally:
            self.file.close()


class FileWriter(TaskRunner):
    def __init__(
        self,
        name: str,
        file: Union[TextIO, IO[bytes]],
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.file = file
        self.chunk_buffer_reader = ChunkBufferReader(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            fd = self.file.fileno()
            while not self.stopped:
                data = self.chunk_buffer_reader.read(0.5)
                if data is not None:
                    if self.stopped or len(data) == 0:
                        return
                    self.offset += len(data)
                    while not self.stopped and len(data) > 0:
                        count = os.write(fd, data)
                        data = data[count:]
        except Exception as e:
            console.print("===Stop write to file===")
            traceback.print_exc()
            console.print("======")
            self.chunk_buffer_reader.close(str(e))
        finally:
            self.file.close()


class BrokerReader(TaskRunner):
    def __init__(
        self,
        name: str,
        broker_url: str,
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.broker_url = broker_url
        self.chunk_buffer_writer = ChunkBufferWriter(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            while not self.stopped:
                data = self.run_until_success(self._read_from_broker)
                if data is not None:
                    if len(data) == 0:
                        self.chunk_buffer_writer.close("")
                        return
                    self.offset += len(data)
                    while not self.stopped and len(data) > 0:
                        count = self.chunk_buffer_writer.write(data, 0.5)
                        data = data[count:]
        except Exception as e:
            console.print("===Stop read from broker===")
            traceback.print_exc()
            console.print("======")
            self.chunk_buffer_writer.close(str(e))
        finally:
            self.run_until_success(self._close_read, True)

    def _read_from_broker(self) -> Optional[bytes]:
        resp = requests.get(self.broker_url, params={"offset": self.offset}, timeout=5)
        check_status_code(resp)
        if "x-wait-for-data" in resp.headers:
            return None
        return resp.content

    def _close_read(self) -> None:
        resp = requests.post(f"{self.broker_url}/closeRead", timeout=5)
        check_status_code(resp)


class BrokerWriter(TaskRunner):
    def __init__(
        self,
        name: str,
        broker_url: str,
        chunk_buffer: ChunkBuffer,
    ):
        super().__init__(name)
        self.broker_url = broker_url
        self.chunk_buffer_reader = ChunkBufferReader(chunk_buffer)
        self.offset = 0

    def run(self) -> None:
        try:
            while not self.stopped:
                data = self.chunk_buffer_reader.read(0.5)
                if data is not None:
                    if self.stopped or len(data) == 0:
                        return
                    d = data
                    while not self.stopped and len(d) > 0:
                        count = self.run_until_success(lambda: self._write_to_broker(d))
                        self.offset += count
                        d = d[count:]
        except Exception as e:
            console.print("===Stop write to broker===")
            traceback.print_exc()
            console.print("======")
            self.chunk_buffer_reader.close(str(e))
        finally:
            self.run_until_success(self._close_write, True)

    def _write_to_broker(self, data: bytes) -> int:
        resp = requests.post(
            self.broker_url, params={"offset": self.offset}, data=data, timeout=5
        )
        check_status_code(resp)
        return resp.json()["count"]  # type: ignore

    def _close_write(self) -> None:
        resp = requests.post(f"{self.broker_url}/closeWrite", timeout=5)
        check_status_code(resp)



import glob
import typing as t
import os.path

from fastapi import APIRouter
from pydantic import Field, BaseModel

from starwhale.api._impl import wrapper
from starwhale.utils.config import SWCliConfigMixed
from starwhale.web.response import success, SuccessResp
from starwhale.api._impl.data_store import SwType, _get_type, TableDesc, LocalDataStore

router = APIRouter()
prefix = "datastore"


class ListTablesRequest(BaseModel):
    prefix: str


class Filter(BaseModel):
    operator: str
    operands: t.List[t.Dict[str, str]]


class QueryTableRequest(BaseModel):
    table_name: str = Field(alias="tableName")
    filter: t.Optional[Filter]
    limit: int
    raw_result: t.Optional[bool] = Field(alias="rawResult")
    encode_with_type: t.Optional[bool] = Field(alias="encodeWithType")


@router.post("/listTables")
def list_tables(request: ListTablesRequest) -> SuccessResp:
    # TODO: use datastore builtin function
    root = str(SWCliConfigMixed().datastore_dir)
    path = os.path.join(root, request.prefix)
    files = glob.glob(f"{path}**", recursive=True)
    tables = []
    for f in files:
        if not os.path.isfile(f):
            continue
        p, file = os.path.split(f)
        p = p[len(root) :].lstrip("/")
        table_name = file.split(".sw-datastore.zip")[0]
        tables.append(f"{p}/{table_name}")
    return success({"tables": tables})


@router.post("/queryTable")
def query_table(request: QueryTableRequest) -> SuccessResp:
    eval_id = _is_eval_summary(request)
    if eval_id:
        return _eval_summary(eval_id)

    ds = LocalDataStore.get_instance()
    rows = list(ds.scan_tables([TableDesc(request.table_name)]))
    if request.encode_with_type:
        return _encode_with_type(rows, request.raw_result or False)
    col, rows = _rows_to_type_and_records(rows)
    return success(
        {
            "columnTypes": col,
            "records": rows,
        }
    )


def _encode_with_type(rows: t.Any, raw_result: bool) -> SuccessResp:
    if not rows:
        return success({"records": [], "columnHints": {}})
    ret = []
    column_hints = {k: SwType.encode_schema(_get_type(v)) for k, v in rows[0].items()}
    for row in rows:
        for k, v in row.items():
            typ = _get_type(v)
            row[k] = typ.encode_type_encoded_value(v, raw_result)
        ret.append(row)
    return success({"records": ret, "columnHints": column_hints})


def _is_eval_summary(request: QueryTableRequest) -> t.Union[str, None]:
    if not request.table_name.endswith("/summary"):
        return None
    if not request.filter:
        return None
    op = request.filter.operator
    operands = request.filter.operands
    if op != "EQUAL":
        return None
    if len(operands) != 2 or {"columnName": "sys/id"} not in operands:
        return None

    eval_id = (
        operands[0].get("intValue")
        if operands[1].get("columnName") == "sys/id"
        else operands[1].get("intValue")
    )

    return eval_id


def _eval_summary(eval_id: str) -> SuccessResp:
    evaluation = wrapper.Evaluation(
        eval_id=eval_id,
        project="self",
        instance="local",
    )
    summary = evaluation.get_metrics()
    col, rows = _rows_to_type_and_records(summary)
    return success(
        {
            "columnTypes": col,
            "records": rows,
        }
    )


def _rows_to_type_and_records(rows: t.Union[list, dict]) -> t.Tuple[list, list]:
    if not rows:
        return [], []
    if not isinstance(rows, list):
        rows = [rows]
    encoders = {k: SwType.encode_schema(_get_type(v)) for k, v in rows[0].items()}
    column_types = [v.update({"name": k}) or v for k, v in encoders.items()]
    return column_types, [{k: str(v) for k, v in row.items()} for row in rows]



from __future__ import annotations

import typing as t
import subprocess
from abc import ABCMeta, abstractmethod

from starwhale.utils import load_yaml
from starwhale.consts import (
    HTTPMethod,
    CREATED_AT_KEY,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
)
from starwhale.utils.fs import move_dir, empty_dir
from starwhale.api._impl import wrapper
from starwhale.base.type import JobOperationType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.http import ignore_error
from starwhale.utils.error import NotFoundError, NoSupportError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.utils.process import check_call
from starwhale.api._impl.metric import MetricKind
from starwhale.base.uri.project import Project
from starwhale.base.uri.resource import Resource

from .store import JobStorage


class Job(metaclass=ABCMeta):
    def __init__(self, uri: Resource) -> None:
        self.uri = uri
        self.name = uri.name
        self.project_name = uri.project.name
        self.sw_config = SWCliConfigMixed()

    @abstractmethod
    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def _get_version(self) -> str:
        raise NotImplementedError

    def _get_report(self) -> t.Dict[str, t.Any]:
        evaluation = wrapper.Evaluation(
            eval_id=self._get_version(),
            project=self.uri.project.name,
            instance=self.uri.instance.url,
        )
        summary = evaluation.get_metrics()
        kind = summary.get("kind", "")

        ret = {
            "kind": kind,
            "summary": summary,
        }

        if kind == MetricKind.MultiClassification.value:
            ret["labels"] = {
                item["id"]: item for item in list(evaluation.get("labels"))
            }
            ret["confusion_matrix"] = {
                "binarylabel": list(evaluation.get("confusion_matrix/binarylabel"))
            }

        return ret

    @classmethod
    def _get_job_cls(
        cls, uri: Project
    ) -> t.Union[t.Type[StandaloneJob], t.Type[CloudJob]]:
        if uri.instance.is_local:
            return StandaloneJob
        elif uri.instance.is_cloud:
            return CloudJob
        else:
            raise NoSupportError(f"job uri:{uri}")

    @classmethod
    def list(
        cls,
        project_uri: Project,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _cls = cls._get_job_cls(project_uri)
        return _cls.list(project_uri)

    @classmethod
    def get_job(cls, job_uri: Resource) -> Job:
        _cls = cls._get_job_cls(job_uri.project)
        return _cls(job_uri)


class StandaloneJob(Job):
    def __init__(self, uri: Resource) -> None:
        super().__init__(uri)
        self.store = JobStorage(uri)

    def _get_version(self) -> str:
        return self.store.id

    @staticmethod
    def _do_flatten_summary(summary: t.Dict[str, t.Any]) -> t.Dict[str, t.Any]:
        rt = {}

        def _f(_s: t.Dict[str, t.Any], _prefix: str = "") -> None:
            for _k, _v in _s.items():
                _k = f"{_prefix}{_k}"
                if isinstance(_v, dict):
                    _f(_v, _prefix=f"{_k}.")
                else:
                    rt[_k] = _v

        _f(summary)
        return rt

    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:
        return {
            "manifest": self.store.manifest,
            "report": self._get_report(),
        }

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        if force:
            empty_dir(self.store.loc)
            return True, ""
        else:
            return move_dir(self.store.loc, self.store.recover_loc, False)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return move_dir(self.store.recover_loc, self.store.loc, force)

    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_container_action(JobOperationType.PAUSE, force)

    def _do_container_action(
        self, action: str, force: bool = False
    ) -> t.Tuple[bool, str]:
        cmd = ["docker"]
        if action == JobOperationType.CANCEL:
            cmd += ["rm", "-fv"]
        elif action == JobOperationType.PAUSE:
            cmd += ["pause"]
        elif action == JobOperationType.RESUME:
            cmd += ["unpause"]

        # search container first
        out = subprocess.check_output(
            ["docker", "ps", "-f", f"label=version={self.store.id}", "-q"]
        )
        _container = out.decode().strip()
        if not _container:
            return (
                False,
                f"failed to {action} container, reason: not found container by label=version={self.store.id}",
            )
        cmd += [_container]
        try:
            check_call(cmd)
        except Exception as e:
            return False, f"failed to {action} container, reason: {e}"
        else:
            return True, f"run {action} successfully"

    @classmethod
    def list(
        cls,
        project_uri: Project,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        _rt = []
        for _path, _is_removed in JobStorage.iter_all_jobs(project_uri):
            _manifest = load_yaml(_path)
            if not _manifest:
                continue

            _rt.append(
                {
                    "location": str(_path.absolute()),
                    "manifest": _manifest,
                    "is_removed": _is_removed,
                }
            )
        return _rt, {}


class CloudJob(Job, CloudRequestMixed):
    def _get_version(self) -> str:
        # TODO:use full version id
        return self.uri.name

    @ignore_error({})
    def info(
        self, page: int = DEFAULT_PAGE_IDX, size: int = DEFAULT_PAGE_SIZE
    ) -> t.Dict[str, t.Any]:
        if not self.uri.project:
            raise NotFoundError("no selected project")
        return {
            "tasks": self._fetch_tasks(page, size),
            "report": self._get_report(),
            "manifest": self._fetch_job_info(),
        }

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support force remove
        return self.do_http_request_simple_ret(
            f"/project/{self.uri.project}/job/{self.name}",
            method=HTTPMethod.DELETE,
            instance=self.uri.instance,
        )

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.RECOVER, force)

    def cancel(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.CANCEL, force)

    def resume(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.RESUME, force)

    def pause(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_job_action(JobOperationType.PAUSE, force)

    def _do_job_action(self, action: str, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support force action
        return self.do_http_request_simple_ret(
            f"/project/{self.uri.project.name}/job/{self.name}/{action}",
            method=HTTPMethod.POST,
            instance=self.uri.instance,
        )

    @classmethod
    @ignore_error(([], {}))
    def list(
        cls,
        project_uri: Project,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        if not project_uri:
            raise NotFoundError("no selected project")
        crm = CloudRequestMixed()

        r = crm.do_http_request(
            f"/project/{project_uri.name}/job",
            params={"pageNum": page, "pageSize": size},
            instance=project_uri.instance,
        ).json()

        jobs = []
        for j in r["data"]["list"]:
            j.pop("owner", None)
            j[CREATED_AT_KEY] = crm.fmt_timestamp(j["createdTime"])
            j["finished_at"] = crm.fmt_timestamp(j["stopTime"])
            j["duration_str"] = crm.fmt_duration(j["duration"])
            jobs.append({"manifest": j})

        return jobs, crm.parse_pager(r)

    @ignore_error({})
    def _fetch_job_info(self) -> t.Dict[str, t.Any]:
        r = self.do_http_request(
            f"/project/{self.project_name}/job/{self.name}",
            instance=self.uri.instance,
        ).json()
        return r["data"]  # type: ignore

    @ignore_error(([], {}))
    def _fetch_tasks(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        r = self.do_http_request(
            f"/project/{self.project_name}/job/{self.name}/task",
            params={"pageNum": page, "pageSize": size},
            instance=self.uri.instance,
        ).json()

        tasks = []
        for _t in r["data"]["list"]:
            _t[CREATED_AT_KEY] = self.fmt_timestamp(_t["createdTime"])  # type: ignore
            tasks.append(_t)

        return tasks, self.parse_pager(r)



from __future__ import annotations

import os
import sys
import shutil
import typing as t
import fnmatch
from abc import ABC
from pathlib import Path

from starwhale.utils import console, pretty_bytes
from starwhale.utils.venv import check_valid_venv_prefix, check_valid_conda_prefix
from starwhale.utils.config import SWCliConfigMixed
from starwhale.base.blob.file import PathLike, BlakeFile, OptionalPathLike

BuiltinPyExcludes = ["__pycache__/", "*.py[cod]", "*$py.class"]


class ObjectStore(ABC):
    def get(self, key: str) -> t.Any:
        raise NotImplementedError

    def put(self, path: PathLike, key: t.Optional[str]) -> t.Any:
        raise NotImplementedError


# TODO: support file lock for concurrent access
class LocalFileStore(ObjectStore):
    def __init__(self, root: OptionalPathLike = None, soft_link: bool = False) -> None:
        if root is None:
            root = SWCliConfigMixed().object_store_dir
        self.root = Path(root)
        self.root.mkdir(parents=True, exist_ok=True)
        self.soft_link = soft_link

    def get(self, key: str) -> t.Union[BlakeFile, None]:
        """
        Get a file from the object store.
        Args:
            key: the hash of the file
        Returns: the BlakeFile object got or None if not found
        """
        if not key:
            raise ValueError("key cannot be empty")
        path = Path(self.root) / key[:2] / key
        if not path.exists():
            return None
        return BlakeFile(path, key)

    def put(self, path: PathLike, key: t.Optional[str] = None) -> BlakeFile:
        """
        Copy a file to the object store.
        Args:
            path: the file to be copied
            key: the hash of the file, if not provided, will be calculated
        Returns: the BlakeFile object putted
        """
        return self._put(path, key)

    def link(self, src: PathLike, key: t.Optional[str] = None) -> None:
        """
        Link a file to the object store. only support hard link for now.
        Args:
            src: the file links from
            key: the hash of the file, if not provided, will be calculated
        """
        file = BlakeFile(src, key)
        dst = Path(self.root) / file.hash[:2] / file.hash
        if dst.exists():
            return
        file.link(dst)

    def _put(self, path: PathLike, key: t.Optional[str] = None) -> BlakeFile:
        file = BlakeFile(path, key)
        dst = Path(self.root) / file.hash[:2] / file.hash
        if not dst.exists():
            dst.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(path, str(dst))
        return BlakeFile(dst, file.hash)

    @staticmethod
    def _is_file_under_dir(file: Path, dir: Path) -> bool:
        """
        Check if the file is under the directory.
        Args:
            file: the file to be checked
            dir: the directory to be checked
        Returns: True if the file is under the directory
        """
        if sys.version_info < (3, 9):
            return str(file).startswith(str(dir))
        return file.is_relative_to(dir)

    @staticmethod
    def _check_if_is_py_env(path: PathLike) -> bool:
        """
        Check if the path is a virtual environment.
        Args:
            path: the path to be checked
        Returns: True if the path is a virtual environment or conda environment
        """
        return check_valid_venv_prefix(path) or check_valid_conda_prefix(path)

    @classmethod
    def _search_all_py_envs(cls, path: PathLike) -> t.List[Path]:
        """
        Search all virtual environments under the path.
        Args:
            path: the path to be searched
        Returns: a list of virtual environments or conda environments
        """
        envs = []
        for root, dirs, files in os.walk(path):
            if cls._check_if_is_py_env(root):
                envs.append(Path(root))
        return envs

    def copy_dir(
        self,
        src_dir: PathLike,
        dst_dir: PathLike,
        excludes: t.List[str],
        ignore_venv_or_conda: bool = True,
    ) -> int:
        """
        Copy a directory from src_dir to the dst_dir using the object store cache.
        When the file is already in the object store, it will be linked instead of copied.
        Args:
            src_dir: the source directory
            dst_dir: the destination directory
            excludes: the files to be excluded
            ignore_venv_or_conda: whether to ignore virtual environments or conda environments
        Returns: total file size in bytes
        """
        src_dir = Path(src_dir)
        dst_dir = Path(dst_dir)
        ignore_dirs = self._search_all_py_envs(src_dir) if ignore_venv_or_conda else []

        size = 0  # no record for soft link (inode not working in windows)

        for src in src_dir.rglob("*"):
            if not src.is_file():
                continue

            # check if the file under any of the ignored dirs
            ignore = False
            for ignore_dir in ignore_dirs:
                if self._is_file_under_dir(src, ignore_dir):
                    ignore = True
                    break
            if ignore or any(
                [fnmatch.fnmatch(str(src), str(src_dir / ex)) for ex in excludes]
            ):
                console.trace(f"ignore to copy file: {src}")
                continue

            file_size = src.stat().st_size
            size += file_size
            console.trace(f"copy file: {src}, size: {pretty_bytes(file_size)}")
            file = self._put(path=src)
            dst = dst_dir / src.relative_to(src_dir)
            dst.parent.mkdir(parents=True, exist_ok=True)
            file.link(dst, soft=self.soft_link)

        return size



import typing as t

import click

from starwhale.utils.cli import AliasedGroup

from .view import (
    JobTermView,
    get_term_view,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    DEFAULT_REPORT_COLS,
)


@click.group(
    "job",
    cls=AliasedGroup,
    help="Job management, list/info/cancel/resume job",
)
@click.pass_context
def job_cmd(ctx: click.Context) -> None:
    ctx.obj = get_term_view(ctx.obj)


@job_cmd.command("list", aliases=["ls"], help="List all jobs in the specified project")
@click.option(
    "-p",
    "--project",
    default="",
    help="Project URI, default is the current selected project.",
)
@click.option("--fullname", is_flag=True, help="Show fullname of swmp version")
@click.option("--show-removed", is_flag=True, help="Show removed dataset")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for job list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for job list"
)
@click.pass_obj
def _list(
    view: t.Type[JobTermView],
    project: str,
    fullname: bool,
    show_removed: bool,
    page: int,
    size: int,
) -> None:
    view.list(
        project, fullname=fullname, show_removed=show_removed, page=page, size=size
    )


@job_cmd.command("remove", aliases=["rm"], help="Remove job")
@click.argument("job")
@click.option(
    "-f",
    "--force",
    is_flag=True,
    help="Force to remove, the removed job cannot recover",
)
def _remove(job: str, force: bool) -> None:
    click.confirm("continue to remove?", abort=True)
    JobTermView(job).remove(force)


@job_cmd.command("recover", help="Recover removed job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to recover")
def _recover(job: str, force: bool) -> None:
    JobTermView(job).recover(force)


@job_cmd.command("pause", help="Pause job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to pause")
def _pause(job: str, force: bool) -> None:
    click.confirm("continue to pause?", abort=True)
    JobTermView(job).pause(force)


@job_cmd.command("resume", help="Resume job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to resume")
def _resume(job: str, force: bool) -> None:
    JobTermView(job).resume(force)


@job_cmd.command("cancel", help="Cancel job")
@click.argument("job")
@click.option("-f", "--force", is_flag=True, help="Force to cancel")
def _cancel(job: str, force: bool) -> None:
    click.confirm("continue to cancel?", abort=True)
    JobTermView(job).cancel(force)


@job_cmd.command("info", help="Inspect job details")
@click.argument("job")
@click.option(
    "--page", type=int, default=DEFAULT_PAGE_IDX, help="Page number for tasks list"
)
@click.option(
    "--size", type=int, default=DEFAULT_PAGE_SIZE, help="Page size for tasks list"
)
@click.option(
    "--max-report-cols",
    type=int,
    default=DEFAULT_REPORT_COLS,
    help="Max table column size for print",
)
@click.option("--web", is_flag=True, help="Open job info page in browser")
@click.pass_obj
def _info(
    view: t.Type[JobTermView],
    job: str,
    page: int,
    size: int,
    max_report_cols: int,
    web: bool,
) -> None:
    view(job).info(page, size, max_report_cols, web)



from __future__ import annotations

import time
import typing as t
from pathlib import Path
from concurrent.futures import as_completed, ThreadPoolExecutor

from starwhale.utils import console
from starwhale.consts import RunStatus
from starwhale.base.context import Context

from .dag import DAG
from .step import Step, StepResult, StepExecutor
from .task import TaskResult, TaskExecutor


class Scheduler:
    def __init__(
        self,
        project: str,
        version: str,
        workdir: Path,
        dataset_uris: t.List[str],
        steps: t.List[Step],
    ) -> None:
        self._steps: t.Dict[str, Step] = {s.name: s for s in steps}
        self.dag: DAG = Step.generate_dag(steps)
        self.project = project
        self.dataset_uris = dataset_uris
        self.workdir = workdir
        self.version = version

    def run(
        self, step_name: str = "", task_num: int = 0, task_index: t.Optional[int] = None
    ) -> t.List[StepResult]:
        if not step_name:
            return self._schedule_all()

        if task_index is None or task_index < 0:
            return [self._schedule_one_step(step_name=step_name, task_num=task_num)]
        else:
            return [self._schedule_one_task(step_name=step_name, task_index=task_index)]

    def _schedule_all(self) -> t.List[StepResult]:
        _results: t.List[StepResult] = []
        # record all vertex's in degree
        indegree_dict = {}
        for vtx in self.dag.vertices():
            indegree_dict[vtx] = self.dag.in_degree(vtx)

        vertices_running: t.Set[str] = set()
        prepared_vertices = self.dag.all_starts()

        while prepared_vertices:
            vertices_to_run = prepared_vertices - vertices_running

            tasks = [
                StepExecutor(
                    self._steps[v],
                    project=self.project,
                    dataset_uris=self.dataset_uris,
                    workdir=self.workdir,
                    version=self.version,
                )
                for v in vertices_to_run
            ]
            with ThreadPoolExecutor(max_workers=len(vertices_to_run)) as pool:
                future_tasks = [pool.submit(t.execute) for t in tasks]
                step_results = [t.result() for t in as_completed(future_tasks)]

            vertices_running |= set(vertices_to_run)
            _results += step_results
            vertices_finished = [result.name for result in step_results]
            vertices_running -= set(vertices_finished)
            prepared_vertices -= set(vertices_finished)

            # update dag
            for result in step_results:
                for v_to in self.dag.successors(result.name):
                    indegree_dict[v_to] -= 1
                    if indegree_dict[v_to] == 0:
                        prepared_vertices.add(v_to)

            if not all(
                [
                    all(tr.status == RunStatus.SUCCESS for tr in sr.task_results)
                    for sr in step_results
                ]
            ):
                break
        return _results

    def _schedule_one_step(self, step_name: str, task_num: int = 0) -> StepResult:
        step = self._steps[step_name]
        start_time = time.time()
        result = StepExecutor(
            step=step,
            project=self.project,
            dataset_uris=self.dataset_uris,
            workdir=self.workdir,
            version=self.version,
            task_num=task_num,
        ).execute()

        console.info(
            f"step:{step_name}, result:{result}, run time:{time.time() - start_time}"
        )
        return result

    def _schedule_one_task(self, step_name: str, task_index: int) -> StepResult:
        _step = self._steps[step_name]
        if task_index >= _step.task_num:
            raise RuntimeError(
                f"task_index:{task_index} out of bounds, total:{_step.task_num}"
            )

        _task = TaskExecutor(
            index=task_index,
            context=Context(
                project=self.project,
                version=self.version,
                step=_step.name,
                total=_step.task_num,
                index=task_index,
                dataset_uris=self.dataset_uris,
                workdir=self.workdir,
            ),
            step=_step,
            workdir=self.workdir,
        )
        start_time = time.time()
        task_result: TaskResult = _task.execute()

        console.info(
            f"step:{step_name}, task result:{task_result}, run time:{time.time() - start_time}"
        )
        return StepResult(name=step_name, task_results=[task_result])



import os
import json
import typing as t
from pathlib import Path

from rich.progress import TaskID, Progress

from starwhale.utils import load_yaml
from starwhale.consts import (
    FileDesc,
    FileNode,
    HTTPMethod,
    SW_BUILT_IN,
    SWMP_SRC_FNAME,
    SW_AUTO_DIRNAME,
    RESOURCE_FILES_NAME,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.utils.fs import extract_tar
from starwhale.utils.retry import http_retry
from starwhale.base.bundle_copy import BundleCopy, _query_param_map
from starwhale.base.uri.instance import Instance
from starwhale.base.uri.resource import Resource, ResourceType


class ModelCopy(BundleCopy):
    def upload_files(self, workdir: Path) -> t.Iterator[FileNode]:
        for _m in self._get_resource_files(workdir):
            if _m["desc"] != FileDesc.MODEL.name:
                continue
            _path = workdir / _m["path"]
            yield FileNode(
                path=_path,
                name=_m["name"],
                size=_path.stat().st_size,
                file_desc=FileDesc.MODEL,
                signature=_m["signature"],
            )

        _meta_names = [SWMP_SRC_FNAME]

        for _n in _meta_names:
            _path = workdir / _n
            yield FileNode(
                path=_path,
                name=os.path.basename(_path),
                size=_path.stat().st_size,
                file_desc=FileDesc.SRC_TAR,
                signature="",
            )

    def _get_resource_files(self, workdir: Path) -> t.List[t.Dict[str, t.Any]]:
        resource_files_path = workdir / "src" / SW_AUTO_DIRNAME / RESOURCE_FILES_NAME
        if resource_files_path.exists():
            resources = load_yaml(resource_files_path)
        else:
            resources = load_yaml(workdir / DEFAULT_MANIFEST_NAME).get("resources", [])
        return resources  # type: ignore

    def download_files(self, workdir: Path) -> t.Iterator[FileNode]:
        for _f in (SWMP_SRC_FNAME,):
            yield FileNode(
                path=workdir / _f,
                signature="",
                size=0,
                name=_f,
                file_desc=FileDesc.SRC_TAR,
            )
            extract_tar(
                tar_path=workdir / SWMP_SRC_FNAME,
                dest_dir=workdir / "src",
                force=False,
            )
        # this must after src download
        for _m in self._get_resource_files(workdir):
            if _m["desc"] != FileDesc.MODEL.name:
                continue
            _sign = _m["signature"]
            _dest = workdir / _m["path"]

            yield FileNode(
                path=_dest,
                signature=_sign,
                size=0,
                name=_m["name"],
                file_desc=FileDesc.MODEL,
            )
            # TODO use unified storage
            # Path(workdir / _m["path"]).symlink_to(
            #     _dest # the unify dir
            # )

    def final_steps(self, progress: Progress) -> None:
        if self.src_uri.instance.is_local:
            manifest_file = (
                self._get_versioned_resource_path(self.src_uri) / DEFAULT_MANIFEST_NAME
            )

            manifest = load_yaml(manifest_file)
            packaged_runtime = manifest.get("packaged_runtime", None)
            if packaged_runtime:
                rt_version = packaged_runtime["manifest"]["version"]
                rt_file_path = (
                    self._get_versioned_resource_path(self.src_uri)
                    / packaged_runtime["path"]
                )
                _tid = progress.add_task(
                    f":arrow_up: synchronize the built-in runtime:{rt_version}..."
                )

                dest_uri = Resource(
                    f"{self.dest_uri.project}/{SW_BUILT_IN}/version/{rt_version}",
                    typ=ResourceType.runtime,
                    refine=False,
                )

                def upload_runtime_tar(file_path: Path, progress: Progress) -> None:
                    task_id = progress.add_task(
                        f":synchronize the built-in runtime {file_path.name}",
                        total=file_path.stat().st_size,
                    )
                    self.do_multipart_upload_file(
                        url_path=f"/project/{dest_uri.project.name}/{ResourceType.runtime.value}/{SW_BUILT_IN}/version/{rt_version}/file",
                        file_path=file_path,
                        instance=dest_uri.instance,
                        fields={
                            _query_param_map[
                                ResourceType.runtime
                            ]: f"{SW_BUILT_IN}:{rt_version}",
                            "project": dest_uri.project.name,
                            "force": "1" if self.force else "0",
                        },
                        use_raise=True,
                        progress=progress,
                        task_id=task_id,
                    )

                upload_runtime_tar(rt_file_path, progress)

                @http_retry
                def sync_built_in_runtime(
                    path: str,
                    instance: Instance,
                    progress: Progress,
                    task_id: TaskID = TaskID(0),
                ) -> None:
                    self.do_http_request(
                        path=path,
                        method=HTTPMethod.PUT,
                        instance=instance,
                        data=json.dumps(
                            {
                                "built_in_runtime": rt_version,
                            }
                        ),
                        use_raise=True,
                        disable_default_content_type=False,
                    )
                    progress.update(task_id, completed=100)

                sync_built_in_runtime(
                    path=self._get_remote_bundle_api_url(for_head=True),
                    instance=self.dest_uri.instance,
                    progress=progress,
                    task_id=_tid,
                )



import threading
from typing import Any, Dict, List, Optional
from datetime import datetime

from fastapi import Body, FastAPI, Request, Response, HTTPException
from pydantic import BaseSettings

from starwhale.utils import console

from .common import (
    random_id,
    ChunkBuffer,
    ClosedError,
    PeriodicRunner,
    InvalidOffsetError,
)


class Settings(BaseSettings):
    broker_url: str = ""
    chunk_size: int = 65536
    max_chunk_count: int = 64
    gc_timeout_seconds = 600.0
    gc_interval_seconds = 60.0
    keep_alive_seconds = 15.0
    keep_alive_check_interval_seconds = 5.0


settings = Settings()
app = FastAPI()
global_lock = threading.Lock()


class Command:
    def __init__(self, args: List[str]):
        self.args = args
        self.exit_code: Optional[int] = None
        self.files: Dict[str, ChunkBuffer] = {
            "stdin": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
            "stdout": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
            "stderr": ChunkBuffer(settings.chunk_size, settings.max_chunk_count),
        }
        self.last_access_time = datetime.now().timestamp()
        self.host_access_time = self.last_access_time
        self.remote_access_time = self.last_access_time

    def get_file(self, filename: str) -> ChunkBuffer:
        if filename not in self.files:
            raise HTTPException(status_code=404, detail=f"invalid filename {filename}")
        return self.files[filename]


class Session:
    def __init__(self) -> None:
        self.commands: Dict[str, Command] = {}
        self.last_access_time = datetime.now().timestamp()

    def new_command(self, command_id: str, args: List[str]) -> None:
        if command_id not in self.commands:
            self.commands[command_id] = Command(args)

    def get_command(self, command_id: str) -> Command:
        if command_id not in self.commands:
            raise HTTPException(
                status_code=404, detail=f"invalid command id {command_id}"
            )
        command = self.commands[command_id]
        command.last_access_time = datetime.now().timestamp()
        return command

    def list_commands(self) -> List[Dict[str, Any]]:
        ret = []
        for command_id, command in self.commands.items():
            ret.append(
                {
                    "command_id": command_id,
                    "args": command.args,
                    "exitCode": command.exit_code,
                }
            )
        return ret


class SessionManager:
    def __init__(self) -> None:
        self.sessions: Dict[str, Session] = {}
        self.garbage_collector: Optional[PeriodicRunner] = None
        self.keep_alive_monitor: Optional[PeriodicRunner] = None

    def keep_alive_check(self) -> None:
        with global_lock:
            now = datetime.now().timestamp()
            for session in self.sessions.values():
                for command in session.commands.values():
                    if now - command.host_access_time > settings.keep_alive_seconds:
                        command.get_file("stdin").close("host timeout")
                        command.get_file("stdout").close("host timeout")
                        command.get_file("stderr").close("host timeout")
                    if now - command.remote_access_time > settings.keep_alive_seconds:
                        command.get_file("stdin").close("remote timeout")
                        command.get_file("stdout").close("remote timeout")
                        command.get_file("stderr").close("remote timeout")

    def garbage_collect(self) -> None:
        with global_lock:
            now = datetime.now().timestamp()
            new_sessions = {}
            for session_id, session in self.sessions.items():
                if now - session.last_access_time > settings.gc_timeout_seconds:
                    console.print(f"garbage collection: remove session {session_id}")
                else:
                    new_sessions[session_id] = session
                    new_commands = {}
                    for command_id, command in session.commands.items():
                        if now - command.last_access_time > settings.gc_timeout_seconds:
                            console.print(
                                f"garbage collection: remove command {command_id}"
                            )
                        else:
                            new_commands[command_id] = command
                    session.commands = new_commands
            self.sessions = new_sessions

    def start_garbage_collector(self) -> None:
        with global_lock:
            if self.garbage_collector is None:
                self.garbage_collector = PeriodicRunner(
                    "broker-garbage-collector",
                    self.garbage_collect,
                    settings.gc_interval_seconds,
                )
                self.garbage_collector.start()
                console.print("garbage collector started")

    def stop_garbage_collector(self) -> None:
        with global_lock:
            if self.garbage_collector is not None:
                self.garbage_collector.stop()
                self.garbage_collector = None

    def start_keep_alive_monitor(self) -> None:
        with global_lock:
            if self.keep_alive_monitor is None:
                self.keep_alive_monitor = PeriodicRunner(
                    "broker-keep-alive-monitor",
                    self.keep_alive_check,
                    settings.keep_alive_check_interval_seconds,
                )
                self.keep_alive_monitor.start()
                console.print("keep alive monitor started")

    def stop_keep_alive_monitor(self) -> None:
        with global_lock:
            if self.keep_alive_monitor is not None:
                self.keep_alive_monitor.stop()
                self.keep_alive_monitor = None

    def new_session(self) -> str:
        session_id = random_id()
        self.sessions[session_id] = Session()
        return session_id

    def get_session(self, session_id: str) -> Session:
        if session_id not in self.sessions:
            raise HTTPException(
                status_code=404, detail=f"invalid session id {session_id}"
            )
        session = self.sessions[session_id]
        session.last_access_time = datetime.now().timestamp()
        return session


session_manager = SessionManager()


@app.on_event("startup")
def startup_event() -> None:
    session_manager.start_garbage_collector()
    session_manager.start_keep_alive_monitor()


@app.post("/session")
def new_session() -> Dict[str, Any]:
    with global_lock:
        session_id = session_manager.new_session()
        return {
            "hostCmd": f"swcli assistance host --broker={settings.broker_url}/session/{session_id}",
            "remoteCmd": f"swcli assistance remote --broker={settings.broker_url}/session/{session_id} [args]",
        }


@app.post("/session/{session_id}/command")
def run_command(
    session_id: str, command_id: str = Body(), args: List[str] = Body()
) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        session.new_command(command_id, args)


@app.post("/session/{session_id}/command/{command_id}")
def update_command_exit_code(session_id: str, command_id: str, exit_code: int) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        command.exit_code = exit_code


@app.post("/session/{session_id}/command/{command_id}/file/{filename}")
async def write_command_file(
    session_id: str,
    command_id: str,
    filename: str,
    offset: int,
    request: Request,
) -> Dict[str, Any]:
    data = await request.body()
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
        if filename == "stdin":
            command.remote_access_time = datetime.now().timestamp()
        else:
            command.host_access_time = datetime.now().timestamp()
    try:
        count = f.write(offset, data, 1)
        return {"count": count}
    except (ClosedError, InvalidOffsetError) as e:
        raise HTTPException(400, str(e))


@app.post("/session/{session_id}/command/{command_id}/file/{filename}/closeRead")
def close_read_command_file(session_id: str, command_id: str, filename: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
    f.close("read close")


@app.post("/session/{session_id}/command/{command_id}/file/{filename}/closeWrite")
def close_write_command_file(session_id: str, command_id: str, filename: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
    f.close("")


@app.get(
    "/session/{session_id}/command/{command_id}/file/{filename}",
    response_class=Response,
    responses={200: {"content": {"application/octet-stream": {}}}},
)
def read_command_file(
    session_id: str, command_id: str, filename: str, offset: int
) -> Response:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        f = command.get_file(filename)
        if filename == "stdin":
            command.host_access_time = datetime.now().timestamp()
        else:
            command.remote_access_time = datetime.now().timestamp()
    try:
        data = f.read(offset, 1)
    except (ClosedError, InvalidOffsetError) as e:
        raise HTTPException(400, str(e))
    if data is None:
        return Response(
            headers={"x-wait-for-data": ""},
            content=b"",
            media_type="application/octet-stream",
        )
    return Response(content=data, media_type="application/octet-stream")


@app.post("/session/{session_id}/command/{command_id}/keepAlive")
def keep_alive(session_id: str, command_id: str, end: str) -> None:
    with global_lock:
        session = session_manager.get_session(session_id)
        command = session.get_command(command_id)
        now = datetime.now().timestamp()
        if end == "host":
            command.host_access_time = now
        else:
            command.remote_access_time = now


@app.get("/session/{session_id}/command")
def list_commands(session_id: str) -> Dict[str, Any]:
    with global_lock:
        session = session_manager.get_session(session_id)
        return {"commands": session.list_commands()}



from __future__ import annotations

import os
import typing as t
from http import HTTPStatus
from pathlib import Path
from concurrent.futures import wait, ThreadPoolExecutor

import yaml
from rich.progress import (
    TaskID,
    Progress,
    BarColumn,
    TextColumn,
    SpinnerColumn,
    TimeElapsedColumn,
    TotalFileSizeColumn,
    TransferSpeedColumn,
)

from starwhale.utils import console, now_str, load_yaml
from starwhale.consts import (
    FileDesc,
    FileNode,
    HTTPMethod,
    CREATED_AT_KEY,
    VERSION_PREFIX_CNT,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.base.tag import StandaloneTag
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.base.type import get_bundle_type_by_uri
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.error import NotFoundError, NoSupportError, FieldTypeOrValueError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.base.blob.store import LocalFileStore
from starwhale.base.uri.project import Project
from starwhale.base.uri.resource import Resource, ResourceType

TMP_FILE_BUFSIZE = 8192

_query_param_map = {
    ResourceType.dataset: "swds",
    ResourceType.model: "swmp",
    ResourceType.runtime: "runtime",
}


class _UploadPhase:
    MANIFEST = "MANIFEST"
    BLOB = "BLOB"
    END = "END"
    CANCEL = "CANCEL"


class BundleCopy(CloudRequestMixed):
    def __init__(
        self,
        src_uri: str | Resource,
        dest_uri: str | Resource | Project,
        typ: ResourceType,
        force: bool = False,
        **kw: t.Any,
    ) -> None:
        self.src_uri: Resource = (
            Resource(src_uri, typ=typ, refine=True)
            if isinstance(src_uri, str)
            else src_uri
        )
        if not self.src_uri.version:
            self.src_uri.version = "latest"

        if not self.src_uri.instance.is_local:
            p = kw.get("dest_local_project_uri")
            project = p and Project(p) or None
            dest_uri = (
                f"{self.src_uri.name}/version/{self.src_uri.version}"
                if (dest_uri == "." or dest_uri == "")
                else dest_uri
            )
        else:
            project = None

        dest = dest_uri
        if isinstance(dest_uri, str):
            try:
                dest = Resource(dest_uri, typ=typ, project=project, refine=False)
            except Exception as e:
                if str(e).startswith("invalid uri"):
                    dest = Project(dest_uri)
                else:
                    raise
        if isinstance(dest, Project):
            self.dest_uri = Resource(
                f"{self.src_uri.name}/version/{self.src_uri.version}",
                name=self.src_uri.name,
                typ=typ,
                project=dest,
                refine=False,
            )
        elif isinstance(dest, Resource):
            self.dest_uri = dest
        else:
            raise Exception("invalid dest_uri")  # this can not happen

        if not self.dest_uri.version or self.dest_uri.version == "latest":
            self.dest_uri.version = self.src_uri.version
        if self.dest_uri.version == "latest":
            self.dest_uri.version = ""

        self.typ = self.src_uri.typ
        self.force = force
        self.field_flag = _query_param_map[self.typ]
        self.field_value = f"{self.src_uri.name}:{self.src_uri.version}"
        self.kw = kw

        self._sw_config = SWCliConfigMixed()
        self._do_validate()
        self._object_store = LocalFileStore()

    def _do_validate(self) -> None:
        if self.typ not in (
            ResourceType.model,
            ResourceType.dataset,
            ResourceType.runtime,
        ):
            raise NoSupportError(f"{self.typ} copy does not work")

    def _check_cloud_obj_existed(self, rc: Resource) -> bool:
        # TODO: add more params for project
        # TODO: tune controller api, use unified params name
        ok, _ = self.do_http_request_simple_ret(
            path=self._get_remote_bundle_api_url(for_head=True),
            method=HTTPMethod.HEAD,
            instance=rc.instance,
            ignore_status_codes=[HTTPStatus.NOT_FOUND],
        )
        return ok

    def _get_versioned_resource_path(self, uri: Resource) -> Path:
        if not uri.instance.is_local:
            raise NoSupportError(f"{uri} to get target dir path")

        return (
            self._sw_config.rootdir
            / uri.project.name
            / self.typ.value
            / uri.name
            / self.src_uri.version[:VERSION_PREFIX_CNT]
            / f"{uri.version}{get_bundle_type_by_uri(uri.typ)}"
        )

    def _check_version_existed(self, uri: Resource) -> bool:
        if uri.instance.is_cloud:
            return self._check_cloud_obj_existed(uri)
        else:
            return self._get_versioned_resource_path(uri).exists()

    def _get_remote_bundle_console_url(self, with_version: bool = True) -> str:
        if self.src_uri.instance.is_cloud:
            remote = self.src_uri
            resource_name = self.src_uri.name
        else:
            remote = self.dest_uri
            resource_name = self.dest_uri.name or self.src_uri.name

        url = f"{remote.instance.url}/projects/{remote.project.name}/{self.typ.value}s/{resource_name}"
        if with_version:
            url = f"{url}/versions/{self.src_uri.version}/overview"
        return url

    def _get_remote_bundle_api_url(self, for_head: bool = False) -> str:
        version = self.src_uri.version
        if not version:
            raise FieldTypeOrValueError(
                f"cannot fetch version from src uri:{self.src_uri}"
            )

        if self.src_uri.instance.is_cloud:
            project = self.src_uri.project
            resource_name = self.src_uri.name
        else:
            project = self.dest_uri.project
            resource_name = self.dest_uri.name or self.src_uri.name

        if not resource_name:
            raise FieldTypeOrValueError(
                f"cannot fetch {self.typ} resource name from src_uri({self.src_uri}) or dest_uri({self.dest_uri})"
            )

        base = [
            f"/project/{project.name}/{self.typ.value}/{resource_name}/version/{version}"
        ]
        if not for_head:
            # uri for head request contains no 'file'
            base.append("file")

        return "/".join(base)

    def _do_upload_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_versioned_resource_path(self.src_uri)
        task_id = progress.add_task(
            f":bowling: upload {file_path.name}",
            total=file_path.stat().st_size,
        )

        self.do_multipart_upload_file(
            url_path=self._get_remote_bundle_api_url(),
            file_path=file_path,
            instance=self.dest_uri.instance,
            fields={
                self.field_flag: self.field_value,
                "project": self.dest_uri.project.name,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )

    def _do_download_bundle_tar(self, progress: Progress) -> None:
        file_path = self._get_versioned_resource_path(self.dest_uri)
        ensure_dir(os.path.dirname(file_path))
        task_id = progress.add_task(f":bowling: download to {file_path}...")

        self.do_download_file(
            url_path=self._get_remote_bundle_api_url(),
            dest_path=file_path,
            instance=self.src_uri.instance,
            params={
                self.field_flag: self.field_value,
                "project": self.src_uri.project.name,
            },
            progress=progress,
            task_id=task_id,
        )

    def do(self) -> None:
        remote_url = self._get_remote_bundle_console_url()
        if not self.force and self._check_version_existed(self.dest_uri):
            console.print(f":tea: {self.dest_uri} was already existed, skip copy")
            return

        if not self._check_version_existed(self.src_uri):
            raise NotFoundError(str(self.src_uri))

        console.print(f":construction: start to copy {self.src_uri} -> {self.dest_uri}")

        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TimeElapsedColumn(),
            TotalFileSizeColumn(),
            TransferSpeedColumn(),
            console=console.rich_console,
            refresh_per_second=0.2,
        ) as progress:
            if self.src_uri.instance.is_local:
                if self.typ == ResourceType.model:
                    self._do_upload_bundle_dir(progress)
                elif self.typ == ResourceType.runtime:
                    self._do_upload_bundle_tar(progress)
                else:
                    raise NoSupportError(
                        f"no support to copy {self.typ} from standalone to server"
                    )
            else:
                if self.typ == ResourceType.model:
                    self._do_download_bundle_dir(progress)
                elif self.typ == ResourceType.runtime:
                    self._do_download_bundle_tar(progress)
                else:
                    raise NoSupportError(
                        f"no support to copy {self.typ} from server to standalone"
                    )
                StandaloneTag(self.dest_uri).add_fast_tag()
                self._update_manifest(
                    self._get_versioned_resource_path(self.dest_uri),
                    {CREATED_AT_KEY: now_str()},
                )
            self.final_steps(progress)
        console.print(f":tea: console url of the remote bundle: {remote_url}")

    def final_steps(self, progress: Progress) -> None:
        pass

    def upload_files(self, workdir: Path) -> t.Iterator[FileNode]:
        raise NotImplementedError

    def download_files(self, workdir: Path) -> t.Iterator[FileNode]:
        raise NotImplementedError

    def _do_download_bundle_dir(self, progress: Progress) -> None:
        _workdir = self._get_versioned_resource_path(self.dest_uri)

        def _download(_tid: TaskID, fd: FileNode) -> None:
            if fd.signature:
                f = self._object_store.get(fd.signature)
                if f is not None and f.exists():
                    f.link(_workdir / fd.path / fd.name)
                    progress.update(_tid, completed=fd.size)
                    return

            self.do_download_file(
                # TODO: use /project/{self.typ}/pull api
                url_path=self._get_remote_bundle_api_url(),
                dest_path=fd.path,
                instance=self.src_uri.instance,
                params={
                    "desc": fd.file_desc.name,
                    "partName": fd.name,
                    "signature": fd.signature,
                },
                progress=progress,
                task_id=_tid,
            )
            # put the downloaded file to object store for cache usage
            self._object_store.link(fd.path, fd.signature)

        _manifest_path = _workdir / DEFAULT_MANIFEST_NAME
        _tid = progress.add_task(f":arrow_down: {DEFAULT_MANIFEST_NAME}")
        _download(
            _tid,
            FileNode(
                path=_manifest_path,
                name=DEFAULT_MANIFEST_NAME,
                signature="",
                file_desc=FileDesc.MANIFEST,
                size=0,
            ),
        )

        for _f in self.download_files(workdir=_workdir):
            _tid = progress.add_task(
                f":arrow_down: {_f.name}",
                total=float(_f.size),
            )
            if not _f.path.exists() or self.force:
                _download(_tid, _f)

    def _do_ubd_bundle_prepare(
        self,
        progress: t.Optional[Progress],
        workdir: Path,
        url_path: str,
    ) -> t.Any:
        manifest_path = workdir / DEFAULT_MANIFEST_NAME
        if progress is None:
            task_id = TaskID(0)
        else:
            task_id = progress.add_task(
                f":arrow_up: {manifest_path.name}",
                total=manifest_path.stat().st_size,
            )

        # TODO: use rich progress
        r = self.do_multipart_upload_file(
            url_path=url_path,
            file_path=manifest_path,
            instance=self.dest_uri.instance,
            params={
                self.field_flag: self.field_value,
                "phase": _UploadPhase.MANIFEST,
                "desc": FileDesc.MANIFEST.name,
                "project": self.dest_uri.project.name,
                "force": "1" if self.force else "0",
            },
            use_raise=True,
            progress=progress,
            task_id=task_id,
        )
        return r.json().get("data", {})

    def _do_ubd_blobs(
        self,
        progress: t.Optional[Progress],
        workdir: Path,
        upload_id: str,
        url_path: str,
        existed_files: t.Optional[t.List] = None,
    ) -> None:
        existed_files = existed_files or []

        # TODO: add retry deco
        def _upload_blob(_tid: TaskID, fd: FileNode) -> None:
            if not fd.path.exists():
                raise NotFoundError(f"{fd.path} not found")

            if progress is not None:
                progress.update(_tid, visible=True)
            self.do_multipart_upload_file(
                url_path=url_path,
                file_path=fd.path,
                instance=self.dest_uri.instance,
                params={
                    self.field_flag: self.field_value,
                    "phase": _UploadPhase.BLOB,
                    "uploadId": upload_id,
                    "partName": fd.name,
                    "signature": fd.signature,
                    "desc": fd.file_desc.name,
                },
                use_raise=True,
                progress=progress,
                task_id=_tid,
            )

        _p_map = {}
        for _id, _f in enumerate(self.upload_files(workdir=workdir)):
            if existed_files and _f.signature in existed_files:
                continue

            if progress is None:
                _tid = TaskID(_id)
            else:
                _tid = progress.add_task(
                    f":arrow_up: {_f.path.name}",
                    total=float(_f.size),
                    visible=False,
                )
            _p_map[_tid] = _f
        with ThreadPoolExecutor(
            max_workers=int(os.environ.get("SW_BUNDLE_COPY_THREAD_NUM", "5"))
        ) as executor:
            futures = [
                executor.submit(_upload_blob, _tid, _file_desc)
                for _tid, _file_desc in _p_map.items()
            ]
            # TODO throw errors
            wait(futures)

    def _do_ubd_end(self, upload_id: str, url_path: str, ok: bool) -> None:
        phase = _UploadPhase.END if ok else _UploadPhase.CANCEL
        self.do_http_request(
            path=url_path,
            method=HTTPMethod.POST,
            instance=self.dest_uri.instance,
            data={
                self.field_flag: self.field_value,
                "project": self.dest_uri.project.name,
                "phase": phase,
                "uploadId": upload_id,
            },
            use_raise=True,
            disable_default_content_type=True,
        )

    def _do_upload_bundle_dir(
        self,
        progress: t.Optional[Progress] = None,
        workdir: t.Optional[Path] = None,
    ) -> None:
        workdir = workdir or self._get_versioned_resource_path(self.src_uri)
        url_path = self._get_remote_bundle_api_url()

        res_data = self._do_ubd_bundle_prepare(
            progress=progress,
            workdir=workdir,
            url_path=url_path,
        )
        upload_id: str = res_data.get("uploadId", "")
        if not upload_id:
            raise Exception("upload id is empty")
        exists_files: list = res_data.get("existed", [])
        try:
            self._do_ubd_blobs(
                progress=progress,
                workdir=workdir,
                upload_id=upload_id,
                url_path=url_path,
                existed_files=exists_files,
            )
        except Exception as e:
            console.print(
                f":confused_face: when upload blobs, we meet Exception{e}, will cancel upload"
            )
            self._do_ubd_end(upload_id=upload_id, url_path=url_path, ok=False)
            raise
        else:
            self._do_ubd_end(upload_id=upload_id, url_path=url_path, ok=True)

    @staticmethod
    def _update_manifest(workdir: Path, patch: t.Dict[str, t.Any]) -> None:
        file = workdir / DEFAULT_MANIFEST_NAME
        # downloaded runtime is a tarball
        if not file.exists():
            return
        manifest = load_yaml(file)
        manifest.update(patch)
        ensure_file(file, yaml.safe_dump(manifest, default_flow_style=False))



from __future__ import annotations

import os
import typing as t
import inspect
import threading
from pathlib import Path

from starwhale.utils import console, disable_progress_bar
from starwhale.utils.fs import blake2b_content
from starwhale.consts.env import SWEnv

from .job import Handler
from ...base.uri.project import Project

_path_T = t.Union[str, Path]
_called_build_functions: t.Dict[str, bool] = {}
_called_build_lock = threading.Lock()


def build(
    modules: t.Optional[t.List[t.Any]] = None,
    workdir: t.Optional[_path_T] = None,
    name: t.Optional[str] = None,
    project_uri: str = "",
    desc: str = "",
    remote_project_uri: t.Optional[str] = None,
    add_all: bool = False,
) -> None:
    """Build Starwhale Model Package.

    In common case, you may call `build` function in your experiment scripts.`build` function is a shortcut for the `swcli model build` command.
    Build function will search all handlers from the `modules` argument or imported modules, and then build Starwhale Model Package.

    Arguments:
        modules: (List[str|object] optional) The search modules supports object(function, class or module) or str(example: "to.path.module", "to.path.module:object").
            If the argument is not specified, the search modules are the imported modules.
        name: (str, optional) The name of Starwhale Model Package, default is the current work dir.
        workdir: (str, Pathlib.Path, optional) The path of the rootdir. The default workdir is the current working dir.
            All files in the workdir will be packaged. If you want to ignore some files, you can add `.swignore` file in the workdir.
        desc: (str, optional) The description of the Starwhale Model Package.
        project_uri: (str, optional) The project uri of the Starwhale Model Package. If the argument is not specified,
            the project_uri is the config value of `swcli project select` command.
        remote_project_uri: (str, optional) The destination project uri(cloud://remote-instance/project/starwhale) of the Starwhale Model Package
        add_all: (bool, optional) Add all files in the workdir to the Starwhale Model Package. If the argument is False, the python cache files and virtualenv files will be ignored.
            the ".swignore" file in the workdir will always take effect.

    Examples:
    ```python
    from starwhale import model

    # class search handlers
    from .user.code.evaluator import ExamplePipelineHandler
    model.build([ExamplePipelineHandler])

    # function search handlers
    from .user.code.evaluator import predict_image
    model.build([predict_image])

    # module handlers, @handler decorates function in this module
    from .user.code import evaluator
    model.build([evaluator])

    # str search handlers
    model.build(["user.code.evaluator:ExamplePipelineHandler"])
    model.build(["user.code1", "user.code2"])

    # no search handlers, use imported modules
    model.build()
    ```

    Returns:
        None.
    """
    from starwhale.core.model.view import ModelTermView
    from starwhale.core.model.model import ModelConfig

    if workdir is None:
        workdir = Path.cwd()
    else:
        workdir = Path(workdir)

    name = name or workdir.name

    search_modules_str = set()
    for h in modules or []:
        if isinstance(h, str):
            search_modules_str.add(h)
        else:
            search_modules_str.add(_ingest_obj_entrypoint_name(h, workdir))

    if not search_modules_str:
        for f in Handler._registered_functions.values():
            search_modules_str.add(_ingest_obj_entrypoint_name(f, workdir))

    if not search_modules_str:
        raise RuntimeError("no modules to search, please specify modules")

    global _called_build_functions, _called_build_lock
    with _called_build_lock:
        arguments = f"{search_modules_str}-{workdir}-{name}-{project_uri}-{desc}-{remote_project_uri}"
        key = blake2b_content(arguments.encode())
        if key in _called_build_functions:
            console.print(
                f":point_right: [bold red]cycle call model build function with the same arguments({arguments}), skip repetitive build"
            )
            return
        else:
            _called_build_functions[key] = True

    with disable_progress_bar():
        Handler.clear_registered_handlers()
        ModelTermView.build(
            workdir=workdir,
            project=project_uri,
            model_config=ModelConfig(
                name=name, run={"modules": list(search_modules_str)}, desc=desc
            ),
            add_all=add_all,
        )

    if remote_project_uri:
        remote_project_uri = Project(remote_project_uri).full_uri
    # TODO support instance and project env in uri component
    elif os.getenv(SWEnv.instance_uri) and os.getenv(SWEnv.project):
        remote_project_uri = (
            f"{os.getenv(SWEnv.instance_uri)}/project/{os.getenv(SWEnv.project)}"
        )

    if remote_project_uri:
        ModelTermView.copy(
            src_uri=f"{Project(project_uri).full_uri}/model/{name}/version/latest",
            dest_uri=remote_project_uri,
        )


def _ingest_obj_entrypoint_name(obj: t.Any, workdir: Path) -> str:
    obj = inspect.unwrap(obj)
    source_path: t.Optional[_path_T] = inspect.getsourcefile(obj)
    if source_path is None:
        raise RuntimeError(f"failed to get source path for object: {obj}")
    source_path = Path(source_path).resolve().absolute()
    workdir = workdir.resolve().absolute()

    relative_path = str(source_path.relative_to(workdir))
    _parts = relative_path.split("/")
    module_import_path = ".".join([p.rsplit(".", 1)[0] for p in _parts if p])
    return module_import_path



from __future__ import annotations

import time
import typing as t
import inspect
import threading
from abc import ABCMeta
from types import TracebackType
from pathlib import Path
from functools import wraps

import dill
import jsonlines

from starwhale.utils import console, now_str
from starwhale.consts import RunStatus, CURRENT_FNAME, DecoratorInjectAttr
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.api._impl import wrapper
from starwhale.base.type import RunSubDirType, PredictLogMode
from starwhale.api.service import Input, Output, Service
from starwhale.utils.error import ParameterError, FieldTypeOrValueError
from starwhale.base.context import Context
from starwhale.core.job.store import JobStorage
from starwhale.api._impl.dataset import Dataset
from starwhale.base.uri.resource import Resource, ResourceType
from starwhale.core.dataset.type import JsonDict
from starwhale.core.dataset.tabular import TabularDatasetRow, TabularDatasetInfo

_jl_writer: t.Callable[[Path], jsonlines.Writer] = lambda p: jsonlines.open(
    str((p).resolve()), mode="w"
)


class PipelineHandler(metaclass=ABCMeta):
    _INPUT_PREFIX = "input/"

    def __init__(
        self,
        predict_batch_size: int = 1,
        ignore_error: bool = False,
        flush_result: bool = False,
        predict_auto_log: bool = True,
        predict_log_mode: str = PredictLogMode.PICKLE.value,
        predict_log_dataset_features: t.Optional[t.List[str]] = None,
        dataset_uris: t.Optional[t.List[str]] = None,
        **kwargs: t.Any,
    ) -> None:
        self.predict_batch_size = predict_batch_size
        self.svc = Service()
        self.context = Context.get_runtime_context()

        self.dataset_uris = self.context.dataset_uris or dataset_uris or []

        self.predict_log_dataset_features = predict_log_dataset_features
        self.ignore_error = ignore_error
        self.flush_result = flush_result
        self.predict_auto_log = predict_auto_log
        self.predict_log_mode = PredictLogMode(predict_log_mode)
        self.kwargs = kwargs

        _logdir = JobStorage.local_run_dir(self.context.project, self.context.version)
        _run_dir = (
            _logdir / RunSubDirType.RUNLOG / self.context.step / str(self.context.index)
        )
        self.status_dir = _run_dir / RunSubDirType.STATUS
        ensure_dir(self.status_dir)

        # TODO: split status/result files
        self._timeline_writer = _jl_writer(self.status_dir / "timeline")

        # TODO: use EvaluationLogStore to refactor this?
        self.evaluation_store = wrapper.Evaluation(
            eval_id=self.context.version, project=self.context.project
        )
        self._update_status(RunStatus.START)

    def __str__(self) -> str:
        return f"PipelineHandler status@{self.status_dir}"

    def __enter__(self) -> PipelineHandler:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        console.debug(f"execute {self.context.step}-{self.context.index} exit func...")
        if value:  # pragma: no cover
            console.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self._timeline_writer.close()

    def _record_status(func):  # type: ignore
        @wraps(func)  # type: ignore
        def _wrapper(*args: t.Any, **kwargs: t.Any) -> None:
            self: PipelineHandler = args[0]
            console.info(
                f"start to run {func.__name__} function@{self.context.step}-{self.context.index} ..."  # type: ignore
            )
            self._update_status(RunStatus.RUNNING)
            try:
                func(*args, **kwargs)  # type: ignore
            except Exception:
                self._update_status(RunStatus.FAILED)
                console.print_exception()
                raise
            else:
                self._update_status(RunStatus.SUCCESS)

        return _wrapper

    @_record_status  # type: ignore
    def _starwhale_internal_run_evaluate(self) -> None:
        now = now_str()
        try:
            self._do_evaluate()
        except Exception as e:
            console.exception(f"evaluate exception: {e}")
            self._timeline_writer.write(
                {"time": now, "status": False, "exception": str(e)}
            )
            raise
        else:
            self._timeline_writer.write({"time": now, "status": True, "exception": ""})

    def _do_predict(
        self,
        data: t.List[t.Dict] | t.Dict,
        index: str | int | t.List[str | int],
        index_with_dataset: str | t.List[str],
        dataset_info: TabularDatasetInfo,
        dataset_uri: Resource,
    ) -> t.Any:
        predict_func = getattr(self, "predict", None)
        ppl_func = getattr(self, "ppl", None)

        if predict_func and ppl_func:
            raise ParameterError("predict and ppl cannot be defined at the same time")

        func = predict_func or ppl_func
        if func is None:
            raise ParameterError(
                "predict or ppl must be defined, predict function is recommended"
            )
        external = {
            "index": index,
            "index_with_dataset": index_with_dataset,
            "dataset_info": dataset_info,
            "context": self.context,
            "dataset_uri": dataset_uri,
        }

        # provide the more flexible way to inject arguments for predict or ppl function
        # case1: only accept data argument
        # 1. def predict(self, data): ...
        # 2. def predict(self, data, /): ...
        # case2: accept data and external arguments
        # 1. def predict(self, *args): ...
        # 2. def predict(self, **kwargs): ...
        # 3. def predict(self, *args, **kwargs): ...
        # 4. def predict(self, data, external: t.Dict): ...
        # 5. def predict(self, data, **kwargs): ...

        kind = inspect._ParameterKind

        parameters = inspect.signature(inspect.unwrap(func)).parameters.copy()
        # Limitation: When the users use custom decorator on the class method(predict/ppl) and only data argument is defined,
        # it is assumed that the first argument related to the class object name is self. Therefore, we remove it from the list of parameters when inspecting them.
        parameters.pop("self", None)

        if len(parameters) <= 0:
            raise RuntimeError("predict/ppl function must have at least one argument")
        elif len(parameters) == 1:
            parameter: inspect.Parameter = list(parameters.values())[0]
            if parameter.kind == kind.VAR_POSITIONAL:
                return func(data, external)
            elif parameter.kind == kind.VAR_KEYWORD:
                return func(data=data, external=external)
            elif parameter.kind in (kind.POSITIONAL_ONLY, kind.POSITIONAL_OR_KEYWORD):
                return func(data)
            else:
                raise RuntimeError(
                    f"unsupported parameter kind for predict/ppl function: {parameter.kind}"
                )
        else:
            return func(data, external=external)

    def _do_evaluate(self) -> t.Any:
        evaluate_func = getattr(self, "evaluate", None)
        cmp_func = getattr(self, "cmp", None)
        if evaluate_func and cmp_func:
            raise ParameterError("evaluate and cmp cannot be defined at the same time")

        func = evaluate_func or cmp_func
        if not func:
            raise ParameterError(
                "evaluate or cmp must be defined, evaluate function is recommended"
            )

        if self.predict_auto_log:
            func(self._iter_predict_result(self.evaluation_store.get_results()))
        else:
            func()

    @_record_status  # type: ignore
    def _starwhale_internal_run_predict(self) -> None:
        if not self.dataset_uris:
            raise FieldTypeOrValueError("context.dataset_uris is empty")
        join_str = "_#@#_"
        cnt = 0
        # TODO: user custom config batch size, max_retries
        for uri_str in self.dataset_uris:
            _uri = Resource(uri_str, typ=ResourceType.dataset)
            ds = Dataset.dataset(_uri, readonly=True)
            ds.make_distributed_consumption(session_id=self.context.version)
            dataset_info = ds.info
            cnt = 0
            if _uri.instance.is_local:
                # avoid confusion with underscores in project names
                idx_prefix = f"{_uri.project.name}/{_uri.name}"
            else:
                r_id = _uri.info().get("id")
                if not r_id:
                    raise KeyError("fetch dataset id error")
                idx_prefix = str(r_id)
            for rows in ds.batch_iter(self.predict_batch_size):
                _start = time.time()
                _exception = None
                _results: t.Any = b""
                try:
                    if self.predict_batch_size > 1:
                        _results = self._do_predict(
                            data=[row.features for row in rows],
                            index=[row.index for row in rows],
                            index_with_dataset=[
                                f"{idx_prefix}{join_str}{row.index}" for row in rows
                            ],
                            dataset_info=dataset_info,
                            dataset_uri=_uri,
                        )
                    else:
                        _results = [
                            self._do_predict(
                                data=rows[0].features,
                                index=rows[0].index,
                                index_with_dataset=f"{idx_prefix}{join_str}{rows[0].index}",
                                dataset_info=dataset_info,
                                dataset_uri=_uri,
                            )
                        ]
                except Exception as e:
                    _exception = e
                    console.exception(
                        f"[{[r.index for r in rows]}] data handle -> failed"
                    )
                    if not self.ignore_error:
                        self._update_status(RunStatus.FAILED)
                        raise
                else:
                    _exception = None

                for (_idx, _features), _result in zip(rows, _results):
                    cnt += 1
                    _idx_with_ds = f"{idx_prefix}{join_str}{_idx}"

                    console.trace(
                        f"[{_idx_with_ds}] use {time.time() - _start:.3f}s, session-id:{self.context.version} @{self.context.step}-{self.context.index}"
                    )

                    self._timeline_writer.write(
                        {
                            "time": now_str(),
                            "status": _exception is None,
                            "exception": str(_exception),
                            "index": _idx,
                            "index_with_dataset": _idx_with_ds,
                        }
                    )

                    self._log_predict_result(
                        features=_features,
                        idx_with_ds=_idx_with_ds,
                        output=_result,
                        idx=_idx,
                    )

        if self.flush_result and self.predict_auto_log:
            self.evaluation_store.flush_result()

        console.info(
            f"{self.context.step}-{self.context.index} handled {cnt} data items for dataset {self.dataset_uris}"
        )

    def _update_status(self, status: str) -> None:
        fpath = self.status_dir / CURRENT_FNAME
        ensure_file(fpath, status)

    def add_api(
        self, input: Input, output: Output, func: t.Callable, name: str
    ) -> None:
        self.svc.add_api(input, output, func, name)

    def serve(self, addr: str, port: int) -> None:
        self.svc.serve(addr, port)

    def _iter_predict_result(
        self, raw_results_iter: t.Iterator[t.Dict]
    ) -> t.Iterator[t.Dict]:
        for data in raw_results_iter:
            mode = data.get("_mode", PredictLogMode.PICKLE.value)
            mode = PredictLogMode(mode)

            if "output" in data and mode == PredictLogMode.PICKLE:
                data["output"] = dill.loads(data["output"])

            input_features = {}
            for k in list(data.keys()):
                if k.startswith(self._INPUT_PREFIX):
                    _, name = k.split(self._INPUT_PREFIX, 1)
                    input_features[name] = data.pop(k)

            data["input"] = input_features
            yield data

    def _log_predict_result(
        self, features: t.Dict, idx_with_ds: str, output: t.Any, idx: str | int
    ) -> None:
        if not self.predict_auto_log:
            return

        if self.predict_log_dataset_features is None:
            _log_features = features
        else:
            _log_features = {
                k: v
                for k, v in features.items()
                if k in self.predict_log_dataset_features
            }

        for artifact in TabularDatasetRow.artifacts_of(_log_features):
            if artifact.link:
                artifact.clear_cache()

        input_features = {
            f"{self._INPUT_PREFIX}{k}": JsonDict.from_data(v)
            for k, v in _log_features.items()
        }
        if self.predict_log_mode == PredictLogMode.PICKLE:
            output = dill.dumps(output)

        # for plain mode: if the output is dict type, we(datastore) will log each key-value pair as a column
        record = {
            "id": idx_with_ds,
            "_mode": self.predict_log_mode.value,
            "_index": idx,
            "output": output,
            **input_features,
        }
        self.evaluation_store.log_result(record)


# TODO: add flush, get_summary functions?
class EvaluationLogStore:
    _instance_holder = threading.local()
    _lock = threading.Lock()

    def __init__(self, id: str, project: str) -> None:
        self.id = id
        self.project = project

        self._datastore = wrapper.Evaluation(eval_id=self.id, project=self.project)

    def __str__(self) -> str:
        return f"Evaluation: id({self.id}), project({self.project})"

    __repr__ = __str__

    @classmethod
    def _get_instance(cls) -> EvaluationLogStore:
        with cls._lock:
            _inst: t.Optional[EvaluationLogStore]
            try:
                _inst = cls._instance_holder.value  # type: ignore
            except AttributeError:
                _inst = None

            if _inst is not None:
                return _inst

            context = Context.get_runtime_context()

            _inst = cls(
                id=context.version,
                project=context.project,
            )

            cls._instance_holder.value = _inst
            return _inst

    @classmethod
    def log(
        cls, category: str, id: t.Union[str, int], metrics: t.Dict[str, t.Any]
    ) -> None:
        """Log metrics into the Starwhale Datastore.

        Arguments:
            category: [str, required] The category of the log records.
            id: [Union[str, int], required] The unique id of the record.
            metrics: [Dict, required] The metrics dict.

        Examples:
        ```python
        from starwhale import evaluation

        evaluation.log("label/1", 1, {"loss": 0.99, "accuracy": 0.98})
        evaluation.log("ppl", "1", {"a": "test", "b": 1})

        Returns:
            None
        ```
        """
        # TODO: support pickle serialize
        el = cls._get_instance()
        el._log(category, id, metrics)

    def _log(
        self, category: str, id: t.Union[str, int], metrics: t.Dict[str, t.Any]
    ) -> None:
        self._datastore.log(table_name=category, id=id, **metrics)

    @classmethod
    def log_summary(cls, *args: t.Any, **kw: t.Any) -> None:
        """Log info into the Starwhale Datastore summary table.

        Arguments:
            *args: [dict, optional] use dicts to store summary info.
            **kwargs: [optional] use kv to store summary info.

        Examples:
        ```python
        from starwhale import evaluation

        evaluation.log_summary(loss=0.99)
        evaluation.log_summary(loss=0.99, accuracy=0.99)
        evaluation.log_summary({"loss": 0.99, "accuracy": 0.99})
        ```

        Returns:
            None
        """
        metrics = {}
        if len(args) == 0:
            metrics = kw
        elif len(args) == 1:
            if len(kw) > 0:
                raise ParameterError(
                    f"Args({args}) and kwargs({kw}) are specified at the same time"
                )
            if not isinstance(args[0], dict):
                raise ParameterError(f"Args({args[0]}) is not dict type")

            metrics = args[0]
        else:
            raise ParameterError(f"The number of args({args}) is greater than one")

        metrics.update(kw)
        el = cls._get_instance()
        el._log_summary(metrics)

    def _log_summary(self, metrics: t.Dict) -> None:
        self._datastore.log_metrics(metrics)

    @classmethod
    def iter(
        cls,
        category: str,
    ) -> t.Iterator:
        """Get an iterator for the log data of the specified category.

        Arguments:
            category: [str, required] The category of the log records.

        Examples:
        ```python
        from starwhale import evaluation

        results = [data for data in evaluation.iter("label/0")]
        ```

        Returns:
            An iterator.
        """
        # TODO: support batch_size
        el = cls._get_instance()
        return el._iter(category)

    def _iter(self, category: str) -> t.Iterator:
        return self._datastore.get(category)


def predict(*args: t.Any, **kw: t.Any) -> t.Any:
    """Defines a predict function.

    This function can be used as a decorator to define an evaluation-predict function that maps the dataset data into
    the different predict workers.

    Arguments:
        datasets: [Union[List[str], None], optional] Use the datasets rows as the input data for the predict function.
        resources: [Dict, optional] Resources for the predict task, such as memory, gpu etc. Current only supports
            the cloud instance.
        concurrency: [int, optional] The concurrency of the predict tasks. Default is 1.
        replicas: [int, optional] The number of the predict tasks. Default is 1.
        batch_size: [int, optional] Number of samples per batch. Default is 1.
        fail_on_error: [bool, optional] Fast fail on the exceptions in the predict function. Default is True.
        auto_log: [bool, optional] Auto log the return values of the predict function and the according dataset rows. Default is True.
        log_mode: [str, optional] When auto_log=True, the log_mode can be specified to control the log behavior. Options are `pickle` and `plain`. Default is `pickle`.
        log_dataset_features: [List[str], optional] When auto_log=True, the log_dataset_features can be specified to control the log dataset features behavior.
            Default is None, all dataset features will be logged. If the list is empty, no dataset features will be logged.
        needs: [List[Callable], optional] The list of the functions that need to be executed before the predict function.

    Examples:
    ```python

    from starwhale import evaluation

    @evaluation.predict
    def predict_image(data):
        ...

    @evaluation.predict(
        dataset="mnist/version/latest",
        batch_size=32,
        replicas=4,
    )
    def predict_batch_images(batch_data)
        ...
    ```

    Returns:
        The decorated function.
    """

    # TODO: support runtime

    if len(args) == 1 and len(kw) == 0 and callable(args[0]):
        return predict()(args[0])
    else:

        def _wrap(func: t.Callable) -> t.Any:
            _register_predict(func, **kw)
            setattr(func, DecoratorInjectAttr.Predict, True)
            return func

        return _wrap


def _register_predict(
    func: t.Callable,
    datasets: t.Optional[t.List[str]] = None,
    resources: t.Optional[t.Dict[str, t.Any]] = None,
    needs: t.Optional[t.List[t.Callable]] = None,
    concurrency: int = 1,
    replicas: int = 1,
    batch_size: int = 1,
    fail_on_error: bool = True,
    auto_log: bool = True,
    log_mode: str = PredictLogMode.PICKLE.value,
    log_dataset_features: t.Optional[t.List[str]] = None,
) -> None:
    from .job import Handler

    Handler.register(
        name="predict",
        resources=resources,
        concurrency=concurrency,
        needs=needs,
        replicas=replicas,
        extra_kwargs=dict(
            predict_batch_size=batch_size,
            ignore_error=not fail_on_error,
            predict_auto_log=auto_log,
            predict_log_mode=log_mode,
            predict_log_dataset_features=log_dataset_features,
            dataset_uris=datasets,
        ),
    )(func)


def evaluate(*args: t.Any, **kw: t.Any) -> t.Any:
    """Defines an evaluate function.

    This function can be used as a decorator to define an evaluation-evaluate function that reduces the results of the
    predict function.

    Argument:
        needs: [List[Callable], required] The list of the functions that need to be executed before the evaluate function.
        use_predict_auto_log: [bool, optional] Passing the iterator of the predict auto-log results into the evaluate function.
            Default is True.
        resources: [Dict, optional] Resources for the predict task, such as memory, gpu etc. Current only supports
            the cloud instance.

    Examples:
    ```python
    from starwhale import evaluation

    @evaluation.evaluate(needs=[predict_image])
    def evaluate_results(predict_result_iter):
        ...

    @evaluation.evaluate(
        use_predict_auto_log=False,
        needs=[predict_image],
    )
    def evaluate_results():
        ...
    ```

    Returns:
        The decorated function.
    """

    def _wrap(func: t.Callable) -> t.Any:
        _register_evaluate(func, **kw)
        setattr(func, DecoratorInjectAttr.Evaluate, True)
        return func

    return _wrap


def _register_evaluate(
    func: t.Callable,
    needs: t.Optional[t.List[t.Callable]] = None,
    resources: t.Optional[t.Dict[str, t.Any]] = None,
    use_predict_auto_log: bool = True,
) -> None:
    from .job import Handler

    if not needs:
        raise ValueError("needs is required for evaluate function")

    Handler.register(
        name="evaluate",
        resources=resources,
        concurrency=1,
        replicas=1,
        needs=needs,
        extra_kwargs=dict(
            predict_auto_log=use_predict_auto_log,
        ),
    )(func)
